package regex ;
import java.util.* ;

class Re {
  static boolean verbose = false ;
  final static int  EMPTY=0, CHAR=1, WILD=2, OR=3, SEQ=4, STAR=5,
    END=6, CHECK=7 /* for backtrack only */ ;
  
  private int tag ;
  private char asChar ;
  private Re p1, p2 ;
  private int pos ;

  private Re() {}

  static Re empty() {
    Re r = new Re() ;
    r.tag = EMPTY ;
    return r ;
  }

  static Re empty = empty() ;

  static Re wild() {
    Re r = new Re() ;
    r.tag = WILD ;
    return r ;
  }

  static Re wild = wild() ;

  static Re charPat(char c) {
    Re r = new Re() ;
    r.tag = CHAR ;
    r.asChar = c ;
    return r ;
  }

  static Re star(Re p) {
    if (p == null) return empty ;
    Re r = new Re () ;
    r.tag = STAR ;
    r.p1 = p ;
    return r ;
  }

  static Re plus(Re p) {
    return seq(p,star(p)) ;
  }

  static Re or(Re p1, Re p2) {
    if (p1 == null) return p2 ;
    if (p2 == null) return p1 ;
    Re r = new Re () ;
    r.tag = OR ;
    r.p1 = p1 ;
    r.p2 = p2 ;
    return r ;
  }

  static Re seq(Re p1, Re p2) {
    if (p1 == null || p2 == null) return null ;
    if (p1 == empty) return p2 ;
    Re r = new Re () ;
    r.tag = SEQ ;
    r.p1 = p1 ;
    r.p2 = p2 ;
    return r ;
  }

  private static void doPrint(Re p, int top, StringBuilder b) {
    if (p == null) b.append("null") ;
    switch (p.tag) {

    case EMPTY:
      b.append("()") ;
      return ;
    case WILD:
      b.append(".") ;
      return ;
    case CHAR:
      switch (p.asChar) {
      case '*': b.append("\\*") ; return ;
      case '+': b.append("\\*") ; return ;
      case '|': b.append("\\|") ; return ;
      case '(': b.append("\\(") ; return ;
      case ')': b.append("\\)") ; return ;
      case '.': b.append("\\.") ; return ;
      default: b.append(p.asChar) ; return ;
      }

    case STAR:
      doPrint(p.p1, STAR,b) ;
      b.append("*") ;
      return ;

    case SEQ:
      if (top == STAR) {
        b.append("(") ;
        doPrint(p.p1, EMPTY, b) ; doPrint(p.p2, EMPTY, b) ;
        b.append(")") ;
      } else {
        doPrint(p.p1, SEQ, b) ; doPrint(p.p2, SEQ, b) ;
      }
      return ;

    case OR:
      if (top == STAR || top == SEQ) {
        b.append("(") ;
        doPrint(p.p1, EMPTY, b) ;
        b.append("|") ;
        doPrint(p.p2, EMPTY, b) ;
        b.append(")") ;
      } else {
        doPrint(p.p1, OR, b) ;
        b.append("|") ;
        doPrint(p.p2, OR, b) ;
      }
      return ;

    case END:
      b.append("$") ;
      return ;
    case CHECK:
      b.append("<"  + p.pos + ",") ;
      doPrint(p.p1, EMPTY, b) ;
      b.append(">") ;
      return ;
    }
    
  }

  public String toString() {
    StringBuilder b = new StringBuilder() ;
    doPrint(this, EMPTY, b) ;
    return b.toString() ;
  }


  static Re parse0(Flux flx) {
    switch (flx.look()) {

    case '\\' :
      flx.next();
      switch (flx.look()) {
      case 'n':
        flx.next();
        return charPat('\n') ;
      case '\\': case '.': case '(': case ')':
      case '*':
        {
          int c = flx.look() ;
          flx.next() ;
          return charPat((char)c) ;
        }
      default:
        flx.syntax("chappement inconnu") ;
        return null ;
      }

    case -1: case ')': case '|':
      return empty() ;
    case '(':
      {
        flx.next() ;
        Re r = regexp(flx) ;
        if (flx.look() != ')') flx.syntax("Parenthse fermante manquante") ;
        flx.next() ;
        return r ;
      }
    case '.':
      flx.next() ;
      return wild() ;

    default:
      {
        int c = flx.look() ;
        flx.next() ;
        return charPat((char)c) ;
      }
    }
  }

  static Re parseStar(Flux flx) {
    Re r = parse0(flx) ;
    while (flx.look() == '*' || flx.look() == '+') {
      r = flx.look() == '*' ? star(r)  : plus(r) ;
      flx.next() ;
    }
    return r ;
  }

  static Re parseSeq(Flux flx) {
    Re r1 = parseStar(flx) ;
    switch(flx.look()) {
    case '|': case ')': case -1:
      return r1 ;
    default:
      return seq(r1, parseSeq(flx)) ;
    }
  }

  static Re parseOr(Flux flx) {
    Re r1 = parseSeq(flx) ;
    switch (flx.look()) {
    case '|':
      flx.next() ;
      return or(r1, parseOr(flx)) ;
    default:
      return r1 ;
    }
  }

  static Re regexp(Flux flx) { return parseOr(flx) ; }

  static Re parse(String t) {
    Flux flx = new Flux(t) ;
    Re r = regexp(flx) ;
    if (!flx.eof()) flx.syntax("fin attendue") ;
    if (verbose) System.err.println( t + " -> " + r) ;
    return r ;
  }

  /* Direct application of filter definition */

  static boolean matches(String text, Re pat, int i, int j) { 
    /*
      System.err.println("  MATCH: "
      + pat + " <" + text.substring(i,j) + ">") ;
    */
    switch (pat.tag) {
      /* Cas de base */
    case EMPTY:
      return i == j ;
    case WILD:
      return i+1 == j ;
    case CHAR:
      return i+1 == j && text.charAt(i) == pat.asChar ;
      /* Induction */
    case OR:
      return matches(text, pat.p1, i, j) || matches(text, pat.p2, i, j) ;
    case SEQ:
      for (int k = i ; k <= j ; k++) {
        if (matches(text, pat.p1, i, k) && matches(text, pat.p2, k, j))
          return true ;
      }
      return false ;
    case STAR:
      if (i == j) {
        return true ;
      } else {
        for (int k = i+1 ; k <= j ; k++) {
          if (matches(text, pat.p1, i, k) && matches(text, pat, k, j))
            return true ;
        }
        return false ;
      }
    }
    throw new Error("Expression regulire inconnue") ;
  }

  /* Derivation */
  static Re derivate(char c, Re p) {
    switch (p.tag) {
    case EMPTY:
      return null ;
    case CHAR:
      return c == p.asChar ? empty : null ;
    case WILD:
      return empty ;
    case OR:
      return or(derivate(c, p.p1), derivate(c, p.p2)) ;
    case SEQ:
      {
        Re q1 = derivate(c, p.p1) ;
        if (nullable(p.p1)) {
          Re q2 = derivate(c, p.p2) ;
          return or(seq(q1, p.p2), q2) ;
        } else {
          return seq(q1, p.p2) ;
        }
      }
    case STAR:
      return seq(derivate(c, p.p1), p) ;
    }
    throw new Error("Expression regulire inconnue") ;
  }

  static boolean nullable(Re p) {
    switch (p.tag) {
    case EMPTY:
      return true ;
    case CHAR:
    case WILD:
      return false ;
    case OR:
      return nullable(p.p1) || nullable(p.p2) ;
    case SEQ:
      return nullable(p.p1) && nullable(p.p2) ;
    case STAR:
      return true ;
    }
    throw new Error("Expression regulire inconnue") ;  
  }

  /* Backtrack */

  // A pattern for end of input
  private static Re end() {
    Re r = new Re() ;
    r.tag = END ;
    return r ;
  }

  private static Re end = end() ;

  // Like seq, except that p1 cannot match the empty word
  private static Re check(int pos, Re p) {
    Re r = new Re () ;
    r.tag = CHECK ;
    r.pos = pos ;
    r.p1 = p ;
    return r ;
  }


  private static class State {
    Re re, kont ;
    int pos ;
    State (Re re, Re kont, int pos) {
      this.re = re ;
      this.kont = kont ;
      this.pos = pos ;
    }
  }

  private static Stack<State> back ;

  private static int scan(Re p,  Re kont, String text, int i) {
    while (p != null) {
      if (verbose) System.err.println("  MATCH: p=" + p + " k=" + kont +
                                    " \"" + text.substring(i) + "\"") ;
      switch (p.tag) {
      case END:
        return i ;
      case EMPTY:
        p = kont ;
        kont = null ;
        break ;
      case WILD:
        if (i < text.length()) {
          p = kont ;
          kont = null ;
          i++ ;
          break ;
        } else
          return -1 ;
      case CHAR:
        if (i < text.length() && p.asChar == text.charAt(i)) {
          p = kont ;
          kont = null ;
          i++ ;
          break ;
        } else {
          return -1 ;
        }
      case STAR:
        back.push(new State(seq(p.p1, check(i,p)), kont, i)) ;
        p = kont ;
        kont = null ;
        break ;
      case OR:
        back.push(new State(p.p2, kont, i)) ;
        p = p.p1 ;
        break ;
      case SEQ:
        Re q = p ;
        p = p.p1 ;
        if (kont != null)
          kont = seq(q.p2, kont) ;
        else
          kont = q.p2 ;
        break ;
      case CHECK:
        if (i == p.pos) return -1 ;
        p = p.p1 ;
        break ;
      default:
        throw new Error("Expression regulire inconnue") ;    
      }
    }
    return -1 ;
  }

  static int back(Re p, String text, int start) {
    if (verbose) System.err.println("PREFIX: <" + start + "> " + p) ;
    back = new Stack<State>() ;
    int nextPos = scan(p, end, text, start) ;
    while (nextPos == -1 && !back.isEmpty()) {
      State s = back.pop() ;
      if (verbose) System.err.println("POP") ;
      nextPos = scan(s.re, s.kont, text, s.pos) ;
    }
    return nextPos ;
  }
}
